/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

import org.xmloperator.reactive.exception.AgentAlreadyLaunchedException;
import org.xmloperator.reactive.exception.BoardAlreadyAttachedException;

/**
 * Community of ReactiveEntity, which are synchronized by instants.
 * Rule : a ReactiveCommunity doesn't publish the reference in memory of
 * 	any ReactiveEntity.
 */
public interface ReactiveCommunity {

	/**
	 * This is the initial status.
	 * During this automata status no ReactiveEntity is connected.
	 * At a connection request, status becomes STATUS_BEFORE_ACTIVE.
	 */
	public static final int STATUS_IDLE = 0;

	/**
	 * This period begins with the proceeding of all the waiting
	 * 	entity connection requests.
	 * If no more entity is connected then the status becomes STATUS_IDLE.
	 * If entities are connected then all the messages (events with attachments)
	 * 	generated at the precedent instant are broadcasted,
	 * 	suspended entities are released and status becomes STATUS_ACTIVE.
	 */
	public static final int STATUS_BEFORE_ACTIVE = 1;
	
	/**
	 * One or more entities are active during this period.
	 * Entities are active during this only status period.
	 * This period ends when no more entity is active.
	 * Next status is STATUS_SUSPENDED.
	 */
	public static final int STATUS_ACTIVE = 2;
	
	/**
	 * This automata status period is used for broadcasting events (with no attachment).
	 * If events occur, next status is STATUS_ACTIVE.
	 * If no event occurs, next status is STATUS_INSTANT_TRANSITION.
	 */
	public static final int STATUS_SUSPENDED = 3;
	
	/**
	 * During this automata status period the instant is incremented.
	 * Next status is STATUS_BEFORE_ACTIVE.
	 */
	public static final int STATUS_INSTANT_TRANSITION = 4;
	
	/**
	 * Returns the current status of this ReactiveCommunity.
	 * @return the current status of this ReactiveCommunity.
	 */
	public int getStatus();
	
	/**
	 * Returns the current instant.
	 * @return the current instant.
	 */
	public long getInstant();
	
	/**
	 * Creates and returns a new base Event.
	 * It is attached to this community.
	 * @return a new base Event.
	 */
	public Event createEvent();

	/**
	 * Creates and returns a new message Event.
	 * It is attached to this community.
	 * @return return a new message Event.
	 */
	public Event createMessageEvent();

	/**
	 * Returns the Action message Event associated to this ReactiveCommunity.
	 * @return the Action message Event associated to this ReactiveCommunity.
	 */
	public Event getActionMessageEvent();
	
	/**
	 * Returns the Board of this ReactiveCommunity.
	 * In order to assume determinism, writing on the Board is realized
	 *   through Action(s) executed during the inactive part of the instants.
	 * Can be null. Becomes final once not null.
	 * @return the Board of this ReactiveCommunity.
	 */
	public Board getBoard();
	
	/**
	 * Set the Board of this ReactiveCommunity.
	 * This is possible only while the Board is null.
	 * The method getBoard() will return the fixed Board at the next instant.
	 * A Board is read-only during the active part of instants.
	 * A Board cannot be attached to more than one ReactiveCommunity.
	 * @param board a Board.
	 * @param permissionKey a permission key for acting on the board.
	 * @exception BoardAlreadyAttachedException if a board is already attached
	 *   to this ReactiveCommunity.
	 */
	public void setBoard(Board board, long permissionKey)
			throws BoardAlreadyAttachedException;

	/**
	 * Launches an Agent into this community.
	 * @param agent : an Agent to launch.
	 * @return the entityId of the launched Agent.
	 * @exception AgentAlreadyLaunchedException if the given Agent is already
	 *   running.
	 */
	public String launchAgent(Agent agent) throws
			AgentAlreadyLaunchedException;

	/**
	 * Returns the connected entity count at this instant.
	 * 
	 * @return the connected entity count at this instant.
	 */
	public int getConnectedEntityCount();
}
