/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.agent;

import org.xmloperator.reactive.Message;

/**
 * A cell is something like a light agent.
 * It belongs to a CellArray, which is an Agent.
 * Usually, cells of a CellArray share the same behavior
 * but have different information contexts.
 */
public abstract class Cell<TCellArray extends CellArray> {
	
	/**
	 * CellArray to which this Cell belongs.
	 */
	protected final TCellArray cellArray;
	
	/**
	 * Number (0..n-1) of this Cell into its CellArray
	 * or -1 if it does'nt operate.
	 */
	protected int cellNumber = -1;
	
	public Cell(TCellArray cellArray) {
		if (cellArray == null)
			throw new IllegalArgumentException();
		this.cellArray = cellArray;
	}
	
	protected final void close() {
		int number = this.cellNumber;
		if (number >= 0)
			this.cellArray.clearCell(number);
	}
	
	/**
	 * Set the number of this Cell. This method is called by the CellArray.
	 * @param cellNumber a Cell number (0..n-1).
	 */
	final void setCellNumber(CellArray cellArray, int cellNumber) {
		if (cellArray != this.cellArray || cellNumber < 0)
			throw new IllegalArgumentException();
		this.cellNumber = cellNumber;
	}

	/**
	 * Clear the number of this Cell. This Cell doesn't operate any more. This
	 * method is called by the CellArray.
	 */
	final void clearCellNumber() {
		this.cellNumber = -1;
	}
	
	/**
	 * Reacts to an exclusive Message.
	 * 
	 * @param message
	 *            an exclusive Message
	 */
	public abstract void reactToMessage(Message message);

	////////////////////////
	// Object implementation
	////////////////////////

	public final String toString() {
		return this.cellArray + "[" + this.cellNumber + "]";
	}
}
