/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.agent;

import java.util.Arrays;

import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.exception.ReactiveException;

/**
 * A CellArray is an array of Cell. There are n positions, numbered 0..count-1.
 * Each position is occupied or not by a Cell. A message destined to a
 * particular Cell is carried by an exclusive Event.
 */
public abstract class CellArray extends ReactiveAgentImpl {
	protected final Cell<?>[] cells;
	
	public CellArray(int cellCount) {
		super(true);
		if (cellCount < 1)
			throw new IllegalArgumentException();
		this.cells = new Cell<?>[cellCount];
		Arrays.fill(this.cells, null);
	}

	public final int getCellCount() {
		return this.cells.length;
	}

	protected final void setCell(int cellNumber, Cell<?> cell) {
		if (this.cells[cellNumber] != null)
			throw new ReactiveException("Cell number already occupied.");
		this.cells[cellNumber] = cell;
		cell.setCellNumber(this, cellNumber);
	}
	
	protected final void clearCell(int cellNumber) {
		Cell<?> cell = this.cells[cellNumber];
		if (cell == null)
			return;
		this.cells[cellNumber] = null;
		cell.clearCellNumber();
	}
	
	///////////////////////////////
	// ReactiveAgent implementation
	///////////////////////////////
	
	public final void reactToCellMessage(Message message) {
		int cellNumber = message.getDestinationCellNumber();
		if (cellNumber >= this.cells.length)
			throw new ReactiveException("Cell number out of bounds");
		this.cells[cellNumber].reactToMessage(message);
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////
	
	public void terminate() {
		super.terminate();
		for (int k = 0; k < this.cells.length; k++)
			this.clearCell(k);
	}
}
