/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.agent;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.exception.ActiveReactiveMachineException;
import org.xmloperator.reactive.exception.DoneMachineException;
import org.xmloperator.reactive.exception.IncorrectKeyException;
import org.xmloperator.reactive.impl.ReactiveCommunityImpl;
import org.xmloperator.reactive.impl.ReactiveEntityImpl;
import org.xmloperator.reactive.ActivityObserver;
import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.CrashingReport;
import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveMachine;
import org.xmloperator.reactive.StartingAndTerminationNotifier;

/**
 * ReactiveMachine implementation.
 */
final class ReactiveMachineImpl
		implements ReactiveMachine {
	private static final long BREATHE_DURATION_MS = 20;

	private static ReactiveMachine currentlyActiveMachine = null;
	private static PrintStream trace = null;
	private static boolean isMachineTrace = false;
	private static boolean isAgentTrace = false;
	
	public static void setTrace(PrintStream trace, boolean isMachineTrace,
			boolean isAgentTrace) {
		ReactiveMachineImpl.trace = trace;
		ReactiveMachineImpl.isMachineTrace = isMachineTrace;
		ReactiveMachineImpl.isAgentTrace = isAgentTrace;
	}

	static ReactiveMachine getCurrentlyActiveMachine() {
		return currentlyActiveMachine;
	}
	
	/**
	 * Gives some time to lower priority threads.
	 */
	private static void breathe() {
		try {
			Thread.sleep(BREATHE_DURATION_MS);
		}
		catch (InterruptedException exception) {}
	}

	private final Object synchroActiveMachine = new Object();
	private final Object synchroMachineTermination = new Object();
	private final long machineDoneKey;
	private final CrashingReport crashingReport;
	private final AtomicInteger aliveAgentCount = new AtomicInteger();
	private final Semaphore launchingSemaphore = new Semaphore(1, true);
	private volatile boolean isAgentLaunched = false;
	private volatile boolean isDone = false;
	private volatile ActivityObserver activityObserver = null;
	
	final StartingAndTerminationNotifier<Agent> agentNotifier =
		new StartingAndTerminationNotifier<Agent>() {

		public void notifyStarting(Agent agent) {
			notifyAgentStarting(agent);
		}

		public void notifyTermination(Agent agent) {
			notifyAgentTermination(agent);
		}
	};

	public ReactiveMachineImpl(CrashingReport report, long machineDoneKey) {
		boolean isActiveMachine;
		synchronized (this.synchroActiveMachine) {
			isActiveMachine = currentlyActiveMachine != null;
			if (!isActiveMachine)
				currentlyActiveMachine = this;
		}
		if (isActiveMachine)
			throw new ActiveReactiveMachineException();
		this.crashingReport = report;
		this.machineDoneKey = machineDoneKey;
	}
	
	/**
	 * Notification of the starting of the given Agent.
	 * @param agent the Agent that is launched.
	 */
	private void notifyAgentStarting(Agent agent) {
		if (trace != null && isMachineTrace)
			trace.println(agent + " starts");
		this.aliveAgentCount.incrementAndGet();
		this.isAgentLaunched = true;
		if (this.activityObserver != null)
			this.activityObserver.notifyAgentStarting(agent);
	}
	
	/**
	 * Notification of the termination of the given Agent.
	 * @param agent the Agent that terminates.
	 */
	private void notifyAgentTermination(Agent agent) {
		if (trace != null && isMachineTrace)
			trace.println(agent + " terminates");
		synchronized (this.synchroMachineTermination) {
			if (this.aliveAgentCount.decrementAndGet() == 0)
				this.synchroMachineTermination.notifyAll();
		}
		if (this.activityObserver != null)
			this.activityObserver.notifyAgentTerminating(agent);
	}
	
	/////////////////////////////////
	// ReactiveMachine implementation
	/////////////////////////////////
	
	public void setActivityObserver(ActivityObserver activityObserver) {
		this.activityObserver = activityObserver;
	}
	
	public ReactiveCommunity newCommunity() {
		return new ReactiveCommunityImpl(this);
	}
	
	public String launchAgent(
			Agent agent,
			ReactiveCommunity reactiveCommunity) {
		if (agent == null || reactiveCommunity == null)
			throw new IllegalArgumentException();
		if (this.isDone)
			throw new DoneMachineException();
		this.launchingSemaphore.acquireUninterruptibly();
		ReactiveEntityImpl reactiveEntity = new ReactiveEntityImpl();
		LaunchingContextImpl launchingContext = new LaunchingContextImpl(
				reactiveEntity, reactiveCommunity, this, isAgentTrace ? trace
						: null);
		new AgentThread(this.agentNotifier, agent, launchingContext,
				this.crashingReport).start();
		Thread.yield();
		while (!this.isAgentLaunched)
			breathe();
		this.isAgentLaunched = false;
		this.launchingSemaphore.release();
		return reactiveEntity.getEntityId();
	}

	public void done(long agentLauncherDoneKey) {
		if (this.isDone)
			return;
		if (agentLauncherDoneKey != this.machineDoneKey)
			throw new IncorrectKeyException();
		/*
		 * Waiting that all agents terminate and all communities has been closed.
		 */
		if (trace != null && isMachineTrace)
			trace.println("ReactiveMachine waits for "
					+ this.aliveAgentCount + " agent(s) be done");
		try {
			synchronized (this.synchroMachineTermination) {
				while (this.aliveAgentCount.get() > 0)
					this.synchroMachineTermination.wait();
			}
		}
		catch (InterruptedException exception) {}
		/*
		 * Terminates the ReactiveMachine.
		 */
		this.isDone = true;
		currentlyActiveMachine = null;
		if (trace != null && isMachineTrace)
			trace.println("ReactiveMachine is done");
	}
}
