/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.ball;

import java.util.List;

import org.xmloperator.reactive.DoubleMessage;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;

import org.xmloperator.reactive.agent.ReactiveAgentImpl;

import org.xmloperator.reactive.impl.DoubleMessageImpl;
import org.xmloperator.reactive.impl.MessageImpl;

/**
 * This Agent plays to pass each received ball to the next player.
 * It aims to test Message processing on exclusive destinations.
 */
public final class BallPlayerAgent2 extends ReactiveAgentImpl {
	private final int playerCount;
	private final int playerNumber;
	private final int previousPlayerNumber;
	private final int nextPlayerNumber;
	private final int stepCount;
	private final Event declarationMessageEvent;
	private final String[] entityIds;
	private final Event[] exclusiveEvents;
	
	public BallPlayerAgent2(int playerCount, int playerNumber, int stepCount,
			Event declarationMessageEvent) {
		super(true);
		if (playerCount < 2 || playerNumber < 0 || stepCount < 0)
			throw new IllegalArgumentException();
		this.playerCount = playerCount;
		this.playerNumber = playerNumber;
		this.previousPlayerNumber = (playerNumber + playerCount - 1) % playerCount;
		this.nextPlayerNumber = (playerNumber + 1) % playerCount;
		this.stepCount = stepCount;
		this.declarationMessageEvent = declarationMessageEvent;
		this.addParameter(playerNumber);
		this.entityIds = new String[playerCount];
		this.exclusiveEvents = new Event[playerCount];
	}

	private final int playerNumber(String identityId) {
		for (int k = 0; k < this.playerCount; k++)
			if (this.entityIds[k] == identityId)
				return k;
		throw new IllegalArgumentException();
	}
	
	private final void passBall(int toPlayerNumber) {
		Event exclusiveEvent = this.exclusiveEvents[toPlayerNumber];
		this.entity.generateMessage(exclusiveEvent, new MessageImpl());
		if (this.trace != null) {
			long instant = this.entity.getCommunity().getInstant();
			this.trace.println(instant + ": " + this.playerNumber + ">"
					+ toPlayerNumber);
		}
	}
	
	///////////////////////////////
	// ReactiveAgent implementation
	///////////////////////////////
	
	/**
	 * Resends the ball to the new destination.
	 */
	public boolean reactToExclusiveMessage(Message message) {
		String senderEntityId = message.getSenderEntityId();
		int senderPlayerNumber = this.playerNumber(senderEntityId);
		if (senderPlayerNumber == this.previousPlayerNumber)
			this.passBall(this.nextPlayerNumber);
		else if (senderPlayerNumber == this.nextPlayerNumber)
			this.passBall(this.previousPlayerNumber);
		else
			throw new IllegalStateException();
		return true;
	}

	public boolean newInstant() {
		long instant = this.entity.getCommunity().getInstant();
		return instant <= this.stepCount + 1;
	}

	///////////////////////
	// Agent implementation
	///////////////////////
	
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		/*
		 * Declares associations between playerNumbers and exclusiveEvents.
		 */
		Event exclusiveEvent = this.entity.getExclusiveMessageEvent();
		this.entity.generateMessage(this.declarationMessageEvent,
				new DoubleMessageImpl(new Integer(this.playerNumber),
						exclusiveEvent));
		this.entity.stop();
		List<Event> eventList = this.entity.suspend();
		if (eventList.size() != 1)
			throw new IllegalStateException();
		Event messageEvent = eventList.get(0);
		if (messageEvent != this.declarationMessageEvent)
			throw new IllegalStateException();
		List<Message> messageList = messageEvent.getMessageList();
		if (messageList.size() != this.playerCount)
			throw new IllegalStateException();
		for (Message message : messageList) {
			DoubleMessage playerMessage =
				(DoubleMessage) message;
			String entityId = playerMessage.getSenderEntityId();
			int player = ((Integer)playerMessage.getObject1()).intValue();
			Event event = ((Event)playerMessage.getObject2());
			if (event.getUniqueDestinationEntityId() != entityId)
				throw new IllegalStateException();
			this.entityIds[player] = entityId;
			this.exclusiveEvents[player] = event;
		}
		/*
		 * The player 0 starts by passing a ball forward and another ball backward.
		 */
		if (this.playerNumber == 0) {
			this.passBall(this.nextPlayerNumber);
			this.passBall(this.previousPlayerNumber);
		}
		this.entity.stop();
	}
}
