/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.agent;

import java.util.ArrayList;
import java.util.List;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.ActionMessage;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.agent.AgentImpl;
import org.xmloperator.reactive.demo.market.ProjectedAccount;
import org.xmloperator.reactive.demo.market.MarketBoard;
import org.xmloperator.reactive.demo.market.Proposal;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.action.OrderAction;
import org.xmloperator.reactive.demo.market.action.DeliveryAction;
import org.xmloperator.reactive.demo.market.action.ProposalMessage;
import org.xmloperator.reactive.exception.ReactiveException;

/**
 * Agent playing the role of an operator in a market place. At instant t, He
 * generates a Proposal. At instant t+1 orders are generated by clients. At
 * instant t+2, this operator receives orders, controls the Proposal and
 * executes deliveries.
 */
public final class Operator extends AgentImpl {
	private final long operatorKey;
	private final Resource proposedResource;
	private final Resource againstResource;
	private final Event proposalMessageEvent;
	private final Event exitMessageEvent;
	private final ResourceSupply mainSupply;
	private final ResourceSupply reciprocalSupply;
	private final ProjectedAccount proposedResourceAccount;
	private final ProjectedAccount againstResourceAccount;
	private boolean mustTerminate = false;
	private double initialValue = 0.0;
	private List<Proposal> proposalList = new ArrayList<Proposal>();
	
	public Operator(MarketBoard marketBoard, long operatorKey,
			ProjectedAccount proposedResourceAccount,
			ProjectedAccount againstResourceAccount,
			Event proposalMessageEvent, Event exitMessageEvent) {
		super(false);
		if (marketBoard == null || proposedResourceAccount == null || againstResourceAccount == null
				|| proposalMessageEvent == null || exitMessageEvent == null)
			throw new IllegalArgumentException();
		this.operatorKey = operatorKey;
		this.proposedResource = proposedResourceAccount.getResource();
		this.againstResource = againstResourceAccount.getResource();
		this.proposedResourceAccount = proposedResourceAccount;
		this.againstResourceAccount = againstResourceAccount;
		this.proposalMessageEvent = proposalMessageEvent;
		this.exitMessageEvent = exitMessageEvent;
		this.mainSupply = new ResourceSupply(marketBoard, this.operatorKey,
				proposedResourceAccount, againstResourceAccount);
		this.reciprocalSupply = new ResourceSupply(marketBoard,
				this.operatorKey, againstResourceAccount,
				proposedResourceAccount);
		this.addParameter(againstResource.getResourceName() + ">"
				+ proposedResource.getResourceName());
	}

	private double getAccountsValue() {
		return this.proposedResourceAccount.getResourceValue(this.operatorKey)
				+ this.againstResourceAccount
						.getResourceValue(this.operatorKey);
	}
	
	private boolean canTerminate() {
		return this.mainSupply.isCampaignListEmpty()
				&& this.reciprocalSupply.isCampaignListEmpty();
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////

	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		this.initialValue = this.getAccountsValue();
		this.mainSupply.setTrace(this.trace);
		this.reciprocalSupply.setTrace(this.trace);
	}
	
	/**
	 * At each instant proposals are generated, orders are processed and
	 * deliveries executed.
	 */
	public void participate() {
		ReactiveCommunity community = this.entity.getCommunity();
		Event actionMessageEvent = community.getActionMessageEvent();
		while (!this.mustTerminate || !this.canTerminate()) {
			long instant = community.getInstant();
			List<ActionMessage> deliveryMessageList = new ArrayList<ActionMessage>();
			/*
			 * Proposals.
			 */
			if (!mustTerminate) {
				Proposal mainProposal = this.mainSupply.startNewCampaign(instant);
				Proposal reciprocalProposal = this.reciprocalSupply.startNewCampaign(instant);
				if (mainProposal != null) {
					this.entity.generateMessage(this.proposalMessageEvent,
							new ProposalMessage(mainProposal));
					this.proposalList.add(mainProposal);
				}
				if (reciprocalProposal != null) {
					this.entity.generateMessage(this.proposalMessageEvent,
							new ProposalMessage(reciprocalProposal));
					this.proposalList.add(reciprocalProposal);
				}
			}
			/*
			 * Listen.
			 */
			List<Event> eventList = this.entity.suspend();
			if (eventList.isEmpty()) {
				mainSupply.removeCampaign(instant - 2);
				reciprocalSupply.removeCampaign(instant - 2);
			}
			else {
				if (eventList.contains(this.exitMessageEvent))
					this.mustTerminate = true;
				/*
				 * Orders reception, storing in supply.orderList.
				 */
				if (eventList.contains(actionMessageEvent))
					for (Message message : actionMessageEvent.getMessageList())
						if (message.isActionMessage()) {
							ActionMessage actionMessage = (ActionMessage) message;
							Action action = actionMessage.getAction();
							Exception exception = actionMessage.getException();
							if (action instanceof OrderAction
									&& exception == null) {
								OrderAction orderAction = (OrderAction) action;
								Resource proposedResource = orderAction.proposal
										.getProposedResource();
								ResourceSupply supply = null;
								if (proposedResource == this.proposedResource)
									supply = this.mainSupply;
								if (proposedResource == this.againstResource)
									supply = this.reciprocalSupply;
								if (supply != null)
									if (!supply.addOrder(orderAction,
											instant - 2)
											&& this.proposalList
													.contains(orderAction.proposal)) {
										/*
										 * Reimbursement.
										 */
										if (this.trace != null)
											this.trace.println(this.entity
													+ " reimburses "
													+ orderAction);
										DeliveryAction deliveryAction = DeliveryAction
												.fullRefund(
														this.operatorKey,
														orderAction,
														supply.againstResource == this.againstResource
																? this.againstResourceAccount
																: this.proposedResourceAccount);
										deliveryMessageList
												.add(this.entity
														.generateActionMessage(deliveryAction));
									}
							}
						}
				/*
				 * Processing supply.orderList and delivery.
				 */
				for (DeliveryAction deliveryAction : this.mainSupply
						.processDelivery(instant - 2))
					deliveryMessageList.add(this.entity
							.generateActionMessage(deliveryAction));
				for (DeliveryAction deliveryAction : this.reciprocalSupply
						.processDelivery(instant - 2))
					deliveryMessageList.add(this.entity
							.generateActionMessage(deliveryAction));
				this.entity.stop();
			}
			/*
			 * Next instant.
			 */
			for (ActionMessage deliveryMessage : deliveryMessageList)
				if (deliveryMessage.getException() != null)
					throw new ReactiveException(deliveryMessage.getException());
		}
		this.entity.stop();
	}

	public void terminate() {
		super.terminate();
		if (this.proposedResourceAccount.hasAnticipations()
				|| this.againstResourceAccount.hasAnticipations())
			throw new IllegalStateException();
		if (this.trace != null && this.initialValue > 0.01) {
			double relativeProfit = this.getAccountsValue() / this.initialValue
					- 1.0;
			trace.println(this.entity + " " + this.againstResource + ">"
					+ this.proposedResource + " profit("
					+ Math.round(this.initialValue) + ">"
					+ Math.round(this.getAccountsValue()) + "): "
					+ Math.round(100.0 * relativeProfit) + "% factor: "
					+ this.mainSupply.getAccumulationFactor() + " reverse: "
					+ this.reciprocalSupply.getAccumulationFactor());
		}
	}
}
