/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.util;

import java.util.Arrays;

/**
 * Real matrix implemented with a single array.
 */
public final class Matrix {

	/**
	 * Returns the product of two Matrices. The number of columns of the first
	 * one must be equal to the number of rows of the second one.
	 * 
	 * @param m1
	 *            the left Matrix.
	 * @param m2
	 *            the right Matrix.
	 * @return the product of the two Matrix.
	 */
	public static Matrix product(Matrix m1, Matrix m2) {
		if (m1.p != m2.n)
			throw new IllegalArgumentException();
		Matrix product = new Matrix(m1.n, m2.p);
		int bp = 0;
		for (int r = 0; r < m1.n; r++)
			for (int c = 0; c < m2.p; c++) {
				double t = 0.0;
				int b1 = r * m1.n;
				int s1 = 1;
				int b2 = c;
				int s2 = m2.p;
				for (int k = 0; k < m1.p; k++) {
					/*
					 * t += m1.terms[r, k] * m2.terms[k, c]
					 */
					t += m1.terms[b1] * m2.terms[b2];
					b1 += s1;
					b2 += s2;
				}
				/*
				 * product.terms[r, c] = t
				 */
				product.terms[bp++] = t;
			}
		return product;
	}

	/**
	 * Returns a new Matrix identity.
	 * @param n the size of the Matrix
	 * @return a Matrix identity with the given size.
	 */
	public static Matrix newIdentity(int n) {
		Matrix m = new Matrix(n, n);
		int b = 0;
		int s = n + 1;
		for (int k = 0; k < n; k++) {
			/*
			 * m.terms[k, k] = 1.0;
			 */
			m.terms[b] = 1.0;
			b += s;
		}
		return m;
	}

	/**
	 * Resolve a linear system of n equations with n unknowns.
	 * Equation at the beginning :
	 * 		A.X = B
	 * Equation at the end :
	 * 		I.X = S
	 * 
	 * @param mat_A
	 *            the matrix A.
	 * @param mat_B
	 *            the vector B.
	 * @return the vector S in case of success, null in case of a null determinant.
	 */
	public static Matrix resolve(Matrix mat_A, Matrix mat_B) {
		Matrix mat_M = mat_A.copy();
		Matrix mat_S = mat_B.copy();
		double[] m = mat_M.terms;
		double[] s = mat_S.terms;
		int n = s.length;
		if (mat_M.n != n || mat_M.p != n || mat_S.p != 1)
			throw new IllegalArgumentException();
		/*
		 * First step : triangulation.
		 */
		for (int row = 0; row < n; row++) {
			/*
			 * Search for a best row.
			 */
			int bestRow = -1;
			{
				double greaterAbsoluteValue = 0.0;
				int b = row;
				for (int r = row; r < n; r++) {
					double absoluteValue = Math.abs(m[b]); // m[r, row]
					b += n;
					if (absoluteValue > greaterAbsoluteValue) {
						bestRow = r;
						greaterAbsoluteValue = absoluteValue;
					}
				}
				if (greaterAbsoluteValue < 1e-100)
					return null;
			}
			/*
			 * Permutes rows 'row' and 'bestRow'.
			 */
			if (bestRow != row) {
				int b1 = row * n;
				int b2 = bestRow * n;
				for (int c = row; c < n; c++) {
					double t = m[b1];
					m[b1] = m[b2];
					m[b2] = t;
					b1++;
					b2++;
				}
				double t = s[row];
				s[row] = s[bestRow];
				s[bestRow] = t;
			}
			/*
			 * Normalizes row 'row'.
			 */
			{
				double pivot = m[row * n + row];
				int b = row * n + row;
				for (int c = row; c < n; c++)
					m[b++] /= pivot; // m[row, c]
				s[row] /= pivot;
			}
			/*
			 * Makes zeroes.
			 */
			for (int r = row + 1; r < n; r++) {
				double f = m[r * n + row];
				int b1 = row * n + row;
				int b2 = r * n + row;
				for (int c = row; c < n; c++)
					m[b2++] -= f * m[b1++]; // m[r, c] -= f * m[row, c];
				s[r] -= f * s[row];
			}
		}
		/*
		 * Second step : diagonalization.
		 */
		for (int row = n - 1; row > 0; row--)
			for (int r = 0; r < row; r++) {
				double f = m[r * n + row];
				int b1 = row * n;
				int b2 = r * n;
				for (int c = 0; c <= row; c++)
					m[b2++] -= f * m[b1++]; // m[r, c] -= f * m[row, c]
				s[r] -= f * s[row];
			}
		/*
		 * Verification : A.S - B =~= 0
		 */
		Matrix mat_E = product(mat_A, mat_S);
		mat_E.subtract(mat_B);
		Matrix mat_Et = mat_E.transpose();
		Matrix mat_E2 = Matrix.product(mat_Et, mat_E);
		if (mat_E2.n != 1 || mat_E2.p != 1)
			throw new IllegalStateException();
		double norm_E = Math.sqrt(mat_E2.get(0, 0) / mat_E.n);
		if (norm_E > 1E-3)
			throw new IllegalStateException();
		return mat_S;
	}
	
	public final int n;
	public final int p;
	
	private final double[] terms;
	
	/**
	 * Constructor of a null Matrix.
	 * @param n the number of row.
	 * @param p the number of columns.
	 */
	public Matrix(int n, int p) {
		if (n < 1 || p < 1)
			throw new IllegalArgumentException();
		this.n = n;
		this.p = p;
		this.terms = new double[n * p];
		Arrays.fill(this.terms, 0.0);
	}

	/**
	 * Transform to the null Matrix.
	 */
	public void clear() {
		Arrays.fill(this.terms, 0.0);
	}
	
	/**
	 * Set this Matrix.
	 * @param terms the terms ordered by row and columns.
	 */
	public void set(double[] terms) {
		if (terms.length != this.terms.length)
			throw new IllegalArgumentException();
		System.arraycopy(terms, 0, this.terms, 0, this.terms.length);
	}
	
	/**
	 * Returns a term of the Matrix.
	 * @param row the row of the term, between 0 and n-1.
	 * @param column the column of the term, between 0 and p-1.
	 * @return the term with the given row and column.
	 */
	public double get(int row, int column) {
		if (row < 0 || row >= n || column < 0 || column >= p)
			throw new IllegalArgumentException();
		return this.terms[row * p + column];
	}
	
	/**
	 * Adds the given Matrix.
	 * @param m a Matrix with the same sizes.
	 */
	public void add(Matrix m) {
		if (m.n != this.n || m.p != this.p)
			throw new IllegalArgumentException();
		int size = this.terms.length;
		for (int k = 0; k < size; k++)
			this.terms[k] += m.terms[k];
	}
	
	/**
	 * Subtracts the given Matrix.
	 * @param m a Matrix with the same sizes.
	 */
	public void subtract(Matrix m) {
		if (m.n != this.n || m.p != this.p)
			throw new IllegalArgumentException();
		int size = this.terms.length;
		for (int k = 0; k < size; k++)
			this.terms[k] -= m.terms[k];
	}

	/**
	 * Returns a copy of this Matrix;
	 * @return a copy of this Matrix;
	 */
	public Matrix copy() {
		Matrix copy = new Matrix(this.n, this.p);
		System.arraycopy(this.terms, 0, copy.terms, 0, this.terms.length);
		return copy;
	}
	
	/**
	 * Returns the transposed Matrix of this.
	 * @return the transposed Matrix of this.
	 */
	public Matrix transpose() {
		Matrix trans = new Matrix(this.p, this.n);
		if (this.n == 1 || this.p == 1)
			System.arraycopy(this.terms, 0, trans.terms, 0, this.terms.length);
		else {
			int b1 = 0;
			int s1 = 1;
			for (int r = 0; r < this.n; r++) {
				int b2 = r;
				int s2 = this.n;
				for (int c = 0; c < this.p; c++) {
					/*
					 * trans.terms[c, r] = this.terms[r, c]
					 */
					trans.terms[b2] = this.terms[b1];
					b1 += s1;
					b2 += s2;
				}
			}
		}
		return trans;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("(");
		int b = 0;
		for (int r = 0; r < this.n; r++) {
			if (r > 0)
				sb.append(",");
			sb.append("(");
			for (int c = 0; c < this.p; c++) {
				if (c > 0)
					sb.append(",");
				sb.append(this.terms[b++]);
			}
			sb.append(")");
		}
		sb.append(")");
		return sb.toString();
	}
}
