/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.syracuse;

import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.agent.AgentImpl;
import org.xmloperator.reactive.impl.SimpleMessageImpl;

/**
 * The purpose of this SyracuseAgent is to stress a ReactiveMachine
 * 	with the Syracuse conjecture.
 * The SyracuseAgent recursively launches new instances of itself.
 */
public final class SyracuseAgent extends AgentImpl {
	private final long start;
	private final long target;
	private Event startMessageEvent;
	
	private long current = 0;
	private boolean isExit = false;

	/**
	 * The Syracuse conjecture recurrence is :
	 *   if n is odd then n -> (3n + 1)/2,
	 *   if n is even then n -> n/2.
	 * The SyracuseAgent does the opposite
	 * 	and generates messages with odd numbers.
	 * @param start a starting strictly positive odd integer.
	 * @param target an odd and strictly positive target.
	 * @param startMessageEvent a MessageEvent to generate.
	 */
	public SyracuseAgent(long start, long target, Event startMessageEvent) {
		super(false);
		if (start < 1 || start % 2 != 1 || target < 1 || target % 2 != 1)
			throw new IllegalArgumentException();
		this.start = start;
		this.target = target;
		this.startMessageEvent = startMessageEvent;
		this.addParameter(start);
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////
	
	/**
	 * Are deterministic :
	 *   the instant count,
	 *   the set of numbers generated within each instant.
	 * Are not deterministic :
	 *   the order of Agent instantiation,
	 *   the relation between entityIds and represented numbers.
	 */
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		this.entity.stop();
		if (this.startMessageEvent == null)
			this.startMessageEvent = launchingContext.getReactiveCommunity()
					.createMessageEvent();
		this.entity.generateMessage(this.startMessageEvent,
				new SimpleMessageImpl(new Long(this.start)));
		if (this.start == this.target) {
			this.isExit = true;
			ExitBoard exitBoard = (ExitBoard) this.entity.getCommunity()
					.getBoard();
			this.entity.generateAction(new ExitAction(exitBoard));
		}
		else if (this.start == 1)
			this.current = 8;
		else {
			long start_mod_3 = this.start % 3; 
			if (start_mod_3 == 0)
				this.isExit = true;
			else {
				this.current = this.start;
				if (start_mod_3 == 1)
					this.current *= 2;
			}
		}
	}

	public void participate() {
		ExitBoard exitBoard = (ExitBoard) this.entity.getCommunity().getBoard();
		while (true) {
			if (this.isExit || exitBoard.isTerminated())
				return;
			{
				long newStart = (2 * this.current - 1) / 3;
				Agent newAgent = new SyracuseAgent(newStart, this.target,
						this.startMessageEvent);
				this.entity.getCommunity().launchAgent(newAgent);
			}
			this.current *= 4;
			this.entity.stop();
		}
	}
}
