/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.exception.BoardAlreadyAttachedException;
import org.xmloperator.reactive.exception.NotOwnEventException;
import org.xmloperator.reactive.exception.ReactiveException;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.Board;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.ReactiveMachine;

/**
 * ReactiveCommunity implementation.
 */
public final class ReactiveCommunityImpl implements ReactiveCommunity {
	
	/**
	 * Group of entities whose threads are released at the same time.
	 */
	class EntityGroup {
		private final boolean isStop;
		private final ReactiveEntityImpl exclusiveDestinationEntity;
		private final Event exclusiveMessageEvent;
		private final AtomicInteger entityCount = new AtomicInteger();
		private final AtomicInteger entityWaitingCount = new AtomicInteger();
		private volatile int entityToWakeUpCount = 0;
		private volatile boolean isGroupReleased = false;

		/**
		 * Constructor for suspended or stopped entities
		 * 
		 * @param isStop
		 *            true for a group of stopped entities, false for a group of
		 *            suspended entities.
		 */
		private EntityGroup(boolean isStop) {
			this.isStop = isStop;
			this.exclusiveDestinationEntity = null;
			this.exclusiveMessageEvent = null;
		}

		/**
		 * Constructor for exclusive destination entity.
		 * 
		 * @param exclusiveDestinationEntity
		 *            an exclusive destination entity.
		 * @param exclusiveMessageEvent
		 *            an exclusive message Event. 
		 */
		EntityGroup(ReactiveEntityImpl exclusiveDestinationEntity,
				Event exclusiveMessageEvent) {
			if (exclusiveDestinationEntity == null
					|| exclusiveMessageEvent == null
					|| exclusiveMessageEvent.getUniqueDestinationEntityId() != exclusiveDestinationEntity
							.getEntityId())
				throw new IllegalArgumentException();
			this.isStop = false;
			this.exclusiveDestinationEntity = exclusiveDestinationEntity;
			this.exclusiveMessageEvent = exclusiveMessageEvent;
		}

		Event getExclusiveMessageEvent() {
			return this.exclusiveMessageEvent;
		}
		
		/**
		 * One entity of this group achieves its cycle. An automata cycle begins
		 * if no more entity (of any group) is active.
		 */
		private final void oneEntityAchieveItsCycle(ReactiveEntity reactiveEntity) {
			/*
			 * Waits for this group wake up.
			 */
			while (this.isGroupReleased && this.entityToWakeUpCount != 0)
				Thread.yield();
			this.isGroupReleased = false;
			/*
			 * Counts the calling entity as belonging to this group.
			 */
			this.entityCount.incrementAndGet();
			if (this.exclusiveDestinationEntity != null) {
				if (reactiveEntity != this.exclusiveDestinationEntity)
					throw new ReactiveException("Not the expected entity");
				synchronized (exclusiveSuspendedEntityList) {
					exclusiveSuspendedEntityList
							.add(this.exclusiveDestinationEntity);
				}
			}
			/*
			 * The calling entity is no more active.
			 */
			if (decrementActivityCount(reactiveEntity) != 0) {
				/*
				 * Others entities (of any group) are active. The calling entity
				 * must wait.
				 */
				this.waitForRelease(reactiveEntity);
				return;
			}
			/*
			 * The activity of every entity of every group is achieved. It is
			 * time for a new cycle. One has to decide the type of the new
			 * cycle.
			 */
			ReactiveEntityImpl exclusiveEntityConcernedByTheNextCycle =
				exclusiveEntityConcernedByTheNextCycle();
			boolean isNewCycleForSuspendedEntities =
				exclusiveEntityConcernedByTheNextCycle == null
					? theNextCyclewillConcernOnlySuspendedEntities()
					: true;
			if (isNewCycleForSuspendedEntities && this.isStop
					|| (exclusiveEntityConcernedByTheNextCycle != null
							&& this.exclusiveDestinationEntity
								!= exclusiveEntityConcernedByTheNextCycle)) {
				/*
				 * Starts a new cycle for suspended entities
				 * (may be an exclusive suspended entity).
				 * The calling entity is not concerned.
				 */
				startNewCycle(exclusiveEntityConcernedByTheNextCycle,
						isNewCycleForSuspendedEntities);
				/*
				 * The calling entity must wait. Others entities are active.
				 */
				this.waitForRelease(reactiveEntity);
			}
			else {
				/*
				 * Counts the calling entity as waiting.
				 */
				this.entityWaitingCount.incrementAndGet();
				/*
				 * Starts a new cycle, which ends with releasing this group.
				 */
				startNewCycle(exclusiveEntityConcernedByTheNextCycle,
						isNewCycleForSuspendedEntities);
				/*
				 * No wait for the calling entity, which is the only active
				 * entity at this time.
				 */
			}
		}

		/**
		 * Waits for release of this group.
		 */
		private final void waitForRelease(ReactiveEntity reactiveEntity) {
			if (this.isGroupReleased)
				throw new ReactiveException(reactiveEntity + " : prematured release.");
			synchronized (this) {
				this.entityWaitingCount.incrementAndGet();
				this.entityToWakeUpCount++;
				while (!this.isGroupReleased) {
					try {
						this.wait();
						/*
						 * Ends by notifyAll in function release().
						 */
					} catch (InterruptedException exception) {}
				}
				/*
				 * Counts the calling entity as waked up.
				 */
				this.entityToWakeUpCount--;
			}
		}
		
		/**
		 * Prepares this group for its release.
		 */
		private boolean prepareRelease() {
			/*
			 * Waits for all entities be waiting.
			 */
			while (true) {
				synchronized (this) {
					if (this.entityWaitingCount.get() == this.entityCount.get())
						break;
				}
				Thread.yield();
			}
			/*
			 * Is there entities within the group ?
			 */
			if (this.entityCount.get() == 0)
				return false;
			/*
			 * Resets the group.
			 */
			this.entityCount.set(0);
			this.entityWaitingCount.set(0);
			if (this.exclusiveDestinationEntity != null)
				synchronized (exclusiveSuspendedEntityList) {
					exclusiveSuspendedEntityList
							.remove(this.exclusiveDestinationEntity);
				}
			/*
			 * Prepares release.
			 */
			this.isGroupReleased = true;
			return true;
		}
		
		/**
		 * Releases all entities of this group for a new cycle.
		 */
		private void release() {
			synchronized (this) {
				this.notifyAll();
			}
		}
	}

	//////////////
	// Static data
	//////////////

	private static final ReactiveEntityImpl[] EMPTY_REACTIVE_ENTITY_ARRAY = new ReactiveEntityImpl[0];
	
	private static AtomicInteger reactiveCommunityCount = new AtomicInteger();
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;

	/////////////////
	// Static methods
	/////////////////

	public static void setTrace(PrintStream trace, boolean isFullTrace) {
		ReactiveCommunityImpl.trace = trace;
		ReactiveCommunityImpl.isFullTrace = isFullTrace;
	}

	private static void startNewInstant(List<Event> generatedMessageEventList,
			boolean generatedMessagesBecomeReadable) {
		for (Event messageEvent : generatedMessageEventList)
			((MessageEventImpl) messageEvent)
					.startNewInstant(generatedMessagesBecomeReadable);
		generatedMessageEventList.clear();
	}
	
	/////////////////////////////
	// Shared final instance data
	/////////////////////////////

	final String communityId = ("C" + String
			.valueOf(ReactiveCommunityImpl.reactiveCommunityCount
					.incrementAndGet())).intern();
	final MessageEventImpl actionMessageEvent;

	////////////////////////
	// Private instance data
	////////////////////////

	/*
	 * Cycle.
	 */
	private volatile int status = STATUS_IDLE;
	private volatile long instant = 0;
	private volatile long cycle = 0;
	/*
	 * Synchronization.
	 */
	private final Object synchroStatus = new Object();
	/*
	 * Entities.
	 */
	private final List<ReactiveEntity> entityToConnectList = new ArrayList<ReactiveEntity>();
	private final Set<ReactiveEntity> connectedEntitySet = new HashSet<ReactiveEntity>();
	private final Map<ReactiveEntity, EntityGroup> fromExclusiveEntityToEntityGroup = new HashMap<ReactiveEntity, EntityGroup>();
	private final List<ReactiveEntityImpl> exclusiveSuspendedEntityList = new ArrayList<ReactiveEntityImpl>();
	private final EntityGroup suspendedEntityGroup = new EntityGroup(false);
	private final EntityGroup stoppedEntityGroup = new EntityGroup(true);
	/*
	 * During agent activity (i.e. automata waiting) we have : activeEntityCount
	 * + exclusiveSuspendedEntityCount + suspendedEntityCount + stoppedEntityCount
	 * == count of connected agent.
	 */
	private final AtomicInteger activeEntityCount = new AtomicInteger();
	/*
	 * Events.
	 */
	private final List<Event> generatedEventList = new ArrayList<Event>();
	private final List<Event> generatedMessageEventList = new ArrayList<Event>();
	private final List<Event> generatedExclusiveMessageEventList = new ArrayList<Event>();
	private final List<Event> usedExclusiveMessageEventList = new ArrayList<Event>();
	private final List<Event> broadcastedEventList = new ArrayList<Event>();
	private final List<Event> unmodifiedBroadcastedEventList = Collections
			.unmodifiableList(this.broadcastedEventList);
	/*
	 * Actions.
	 */
	private final List<Action> executableActionList = new ArrayList<Action>();
	/*
	 * Board
	 */
	private volatile Board futureBoard = null;
	private volatile Board board = null;
	private volatile long boardPermissionKey = 0;
	/*
	 * ReactiveMachine.
	 */
	private final ReactiveMachine reactiveMachine;
	
	/**
	 * Constructor.
	 */
	public ReactiveCommunityImpl(ReactiveMachine reactiveMachine) {
		if (reactiveMachine == null)
			throw new IllegalArgumentException();
		this.reactiveMachine = reactiveMachine;
		this.actionMessageEvent = new MessageEventImpl(this);
	}

	/**
	 * Returns the first entity that is exclusively suspended and for which an
	 * exclusive event is destined. The returned Event (if any) and any
	 * exclusive event that is not usable is removed from the list.
	 * 
	 * @return the first entity that is exclusively suspended and for which an
	 *         exclusive event is destined.
	 */
	private ReactiveEntityImpl exclusiveEntityConcernedByTheNextCycle() {
		while (!this.usedExclusiveMessageEventList.isEmpty()) {
			EventImpl event = (EventImpl) this.usedExclusiveMessageEventList.remove(0);
			ReactiveEntityImpl exclusiveEntity = event.uniqueDestinationEntity;
			if (exclusiveEntity == null)
				throw new IllegalStateException();
			ReactiveCommunityImpl.EntityGroup entityGroup = exclusiveEntity
					.getExclusiveEntityGroup();
			if (entityGroup != null && entityGroup.entityCount.get() != 0)
				return exclusiveEntity;
		}
		return null;
	}

	/**
	 * Returns true if the next cycle will occur in the same instant because
	 * there are suspended entities and instantaneous events to broadcast.
	 * Return false if the next cycle will concern a new instant. This method
	 * may only be called if <tt>exclusiveEntityConcernedByTheNextCycle</tt>
	 * returns null.
	 * 
	 * @return true if the next cycle will occur in the same instant because
	 *         there are suspended entities and instantaneous events to
	 *         broadcast, false elsewhere.
	 * @throws IllegalStateException
	 *             if there is waiting exclusive events.
	 */
	private boolean theNextCyclewillConcernOnlySuspendedEntities() {
		if (!this.usedExclusiveMessageEventList.isEmpty())
			throw new IllegalStateException();
		return (this.suspendedEntityGroup.entityCount.get() != 0
				|| !this.exclusiveSuspendedEntityList.isEmpty())
			&& !this.generatedEventList.isEmpty();
	}

	/**
	 * Starts a new cycle automata.
	 */
	private void startNewCycle() {
		ReactiveEntityImpl exclusiveEntityConcernedByTheNextCycle = this
				.exclusiveEntityConcernedByTheNextCycle();
		boolean theNextCyclewillConcernOnlySuspendedEntities =
			exclusiveEntityConcernedByTheNextCycle == null
				? this.theNextCyclewillConcernOnlySuspendedEntities()
				: true;
		this.startNewCycle(exclusiveEntityConcernedByTheNextCycle,
				theNextCyclewillConcernOnlySuspendedEntities);
	}

	/**
	 * Starts a new automata cycle.
	 * 
	 * @param exclusiveEntityConcernedByTheNextCycle
	 *            an exclusive entity concerned by the next cycle, if any, null
	 *            elsewhere.
	 * @param isNewCycleForSuspendedEntities
	 *            if true then starts a cycle for suspended entities else starts
	 *            a cycle for all entities. This argument is only taken into
	 *            account if the previous one is null.
	 */
	private void startNewCycle(
			ReactiveEntityImpl exclusiveEntityConcernedByTheNextCycle,
			boolean isNewCycleForSuspendedEntities) {
		/*
		 * Status: ACTIVE -> SUSPENDED
		 */
		if (this.status != STATUS_ACTIVE)
			throw new ReactiveException("Automata status should be ACTIVE.");
		this.status = STATUS_SUSPENDED;
		/*
		 * Verify no more entity activity.
		 */
		if (this.activeEntityCount.get() != 0)
			throw new ReactiveException("There should be no active entity.");
		/*
		 * Verify:
		 * connectedEntities == exclusiveSuspendedEntities + suspendedEntities + stoppedEntities.
		 */
		if (this.exclusiveSuspendedEntityList.size()
				+ this.suspendedEntityGroup.entityCount.get()
				+ this.stoppedEntityGroup.entityCount.get()
				!= this.connectedEntitySet.size())
			throw new ReactiveException(
					"Connected entities should be suspended or stopped.");
		/*
		 * New Cycle.
		 */
		this.cycle++;
		/*
		 * Starting the new cycle depends on the concerned group(s) of entities.
		 */
		if (exclusiveEntityConcernedByTheNextCycle != null)
			this.startNewCycleForAnExclusiveSuspendedEntity(
					exclusiveEntityConcernedByTheNextCycle);
		else if (isNewCycleForSuspendedEntities)
			this.startNewCycleForTheSuspendedEntities();
		else
			this.startNewInstant();
	}

	private void startNewCycleForAnExclusiveSuspendedEntity(
			ReactiveEntityImpl exclusiveEntityConcernedByTheNextCycle) {
		Event exclusiveMessageEvent = exclusiveEntityConcernedByTheNextCycle
				.getExclusiveMessageEvent();
		EntityGroup entityGroup = exclusiveEntityConcernedByTheNextCycle
				.getExclusiveEntityGroup();
		/*
		 * Broadcasts exclusive event.
		 */
		this.broadcastedEventList.clear();
		this.broadcastedEventList.add(exclusiveMessageEvent);
		/*
		 * Exclusive suspended entity becomes active again.
		 */
		this.activeEntityCount.set(1);
		/*
		 * Prepares release of the group.
		 */
		if (!entityGroup.prepareRelease())
			throw new ReactiveException("The exclusive entity group should contains one entity.");
		/*
		 * Status: SUSPENDED -> ACTIVE
		 */
		this.status = STATUS_ACTIVE;
		/*
		 * Release exclusive suspended entity.
		 */
		entityGroup.release();
	}
	
	private void startNewCycleForTheSuspendedEntities() {
		/*
		 * Broadcasts events.
		 */
		this.broadcastedEventList.clear();
		this.broadcastedEventList.addAll(this.generatedEventList);
		this.generatedEventList.clear();
		/*
		 * Suspended entities become active again.
		 */
		this.activeEntityCount.set(this.exclusiveSuspendedEntityList.size()
				+ this.suspendedEntityGroup.entityCount.get());
		/*
		 * Prepares release of the suspended entity groups.
		 */
		ReactiveEntityImpl[] remainingExclusiveSuspendedEntities =
			this.prepareRemainingExclusiveSuspendedEntitiesRelease();
		this.suspendedEntityGroup.prepareRelease();
		/*
		 * Status: SUSPENDED -> ACTIVE
		 */
		this.status = STATUS_ACTIVE;
		/*
		 * Release suspended entities.
		 */
		for (ReactiveEntityImpl entity : remainingExclusiveSuspendedEntities)
			entity.getExclusiveEntityGroup().release();
		this.suspendedEntityGroup.release();
	}
	
	private void startNewInstant() {
		/*
		 * For this instant, potentials events are lost, since there is no
		 * destination entity.
		 */
		this.broadcastedEventList.clear();
		this.generatedEventList.clear();
		/*
		 * Status: SUSPEND -> INSTANT_TRANSITION -> BEFORE_ACTIVE
		 */
		this.status = STATUS_INSTANT_TRANSITION;
		this.instant += 1;
		if (trace != null)
			trace.nextInstant(this);
		this.status = STATUS_BEFORE_ACTIVE;
		/*
		 * Validates board, executes actions, processes action messages.
		 */
		this.board = this.futureBoard;
		this.executeActions();
		this.processActionMessages();
		/*
		 * Connects requesting entities.
		 */
		while (!this.entityToConnectList.isEmpty()) {
			ReactiveEntity reactiveEntity;
			synchronized (this.entityToConnectList) {
				reactiveEntity = (ReactiveEntity) this.entityToConnectList.remove(0);
			}
			this.connectedEntitySet.add(reactiveEntity);
		}
		/*
		 * Connected entities become active.
		 */
		this.activeEntityCount.set(this.connectedEntitySet.size());
		/*
		 * Is there connected entities ?
		 */
		if (this.activeEntityCount.get() == 0) {
			/*
			 * There is no connected entity. Eventual message events are lost.
			 */
			startNewInstant(this.generatedMessageEventList, false);
			startNewInstant(this.generatedExclusiveMessageEventList, false);
			/*
			 * Status: BEFORE_ACTIVE -> IDLE
			 */
			this.status = STATUS_IDLE;
			if (trace != null)
				trace.nextInstant(null);
		} else {
			/*
			 * There are connected entities. Message events (if any) are moved in order to
			 * be broadcasted during the first suspend cycle of this instant.
			 */
			this.generatedEventList.addAll(this.generatedMessageEventList);
			this.usedExclusiveMessageEventList.addAll(this.generatedExclusiveMessageEventList);
			startNewInstant(this.generatedMessageEventList, true);
			startNewInstant(this.generatedExclusiveMessageEventList, true);
			/*
			 * Prepares release of the entities groups.
			 */
			ReactiveEntityImpl[] remainingExclusiveSuspendedEntities =
				this.prepareRemainingExclusiveSuspendedEntitiesRelease();
			this.suspendedEntityGroup.prepareRelease();
			this.stoppedEntityGroup.prepareRelease();
			/*
			 * Status: BEFORE_ACTIVE -> ACTIVE
			 */
			this.status = STATUS_ACTIVE;
			/*
			 * Releases all entities. No Event is return.
			 */
			for (ReactiveEntityImpl entity : remainingExclusiveSuspendedEntities)
				entity.getExclusiveEntityGroup().release();
			this.suspendedEntityGroup.release();
			this.stoppedEntityGroup.release();
		}
	}
	
	private ReactiveEntityImpl[] prepareRemainingExclusiveSuspendedEntitiesRelease() {
		ReactiveEntityImpl[] remainingExclusiveSuspendedEntities =
			this.exclusiveSuspendedEntityList
				.toArray(EMPTY_REACTIVE_ENTITY_ARRAY);
		for (ReactiveEntityImpl entity : remainingExclusiveSuspendedEntities)
			entity.getExclusiveEntityGroup().prepareRelease();
		if (!this.exclusiveSuspendedEntityList.isEmpty())
			throw new ReactiveException(
					"The exclusive suspended entity list should be empty.");
		return remainingExclusiveSuspendedEntities;
	}
	
	/**
	 * Decrements the active entity count.
	 * 
	 * @return the new count.
	 */
	private int decrementActivityCount(ReactiveEntity reactiveEntity) {
		int activityCount = this.activeEntityCount.decrementAndGet();
		if (activityCount < 0)
			throw new ReactiveException("Active entity count cannot be negative.");
		return activityCount;
	}

	private void executeActions() {
		if (this.executableActionList.isEmpty())
			return;
		for (int priority = Action.PRIORITY_HIGH; priority >= Action.PRIORITY_LOW; priority--)
			for (Action action : this.executableActionList)
				if (action.getPriority() == priority) {
					try {
						action.execute(this.boardPermissionKey);
					} catch (Exception exception) {
						this.reactOnException(action, exception);
					}
				}
		this.executableActionList.clear();
	}
	
	private void reactOnException(Action action, Exception exception) {
		Event reactionEvent = action.getReactionEvent();
		if (reactionEvent != null) {
			ReactiveEntity destinationEntity = ((EventImpl) reactionEvent).uniqueDestinationEntity;
			ActionMessageImpl actionMessage = new ActionMessageImpl(action);
			actionMessage.setException(exception);
			actionMessage.setSenderEntityId(destinationEntity.getEntityId());
			this.generateMessage(destinationEntity, reactionEvent,
					actionMessage);
		}
	}
	
	private void processActionMessages() {
		List<Message> messageList = this.actionMessageEvent
				.getGeneratedMessageList();
		if (messageList.isEmpty())
			return;
		for (int priority = Action.PRIORITY_HIGH; priority >= Action.PRIORITY_LOW; priority--)
			for (Message message : messageList)
				if (message.isActionMessage()) {
					ActionMessageImpl actionMessage = (ActionMessageImpl) message;
					Action action = actionMessage.getAction();
					if (action.getPriority() == priority) {
						try {
							action.execute(this.boardPermissionKey);
						} catch (Exception exception) {
							actionMessage.setException(exception);
							this.reactOnException(action, exception);
						}
					}
				}
	}

	////////////////////////////////////////////
	// Methods to be called by ReactiveEntityImpl
	////////////////////////////////////////////

	/**
	 * Connects a given ReactiveEntity.
	 * @param reactiveEntity an entity to connect to this community.
	 */
	void connect(ReactiveEntity reactiveEntity) {
		synchronized (this.entityToConnectList) {
			this.entityToConnectList.add(reactiveEntity);
		}
		synchronized (this.synchroStatus) {
			if (this.status == STATUS_IDLE) {
				this.status = STATUS_BEFORE_ACTIVE;
				this.entityToConnectList.remove(reactiveEntity);
				this.connectedEntitySet.add(reactiveEntity);
				this.activeEntityCount.incrementAndGet();
				if (trace != null)
					trace.nextInstant(this);
				this.status = STATUS_ACTIVE;
				return;
			}
		}
		while (!this.connectedEntitySet.contains(reactiveEntity));
			Thread.yield();
		while (this.getStatus() != ReactiveCommunity.STATUS_ACTIVE)
			Thread.yield();
	}

	/**
	 * Generates an Event to this community.
	 * @param reactiveEntity the sender reactiveEntity.
	 * @param event a base Event.
	 */
	void generateEvent(ReactiveEntity reactiveEntity, Event event) {
		if (event.getCommunity() != this)
			throw new NotOwnEventException();
		if (((EventImpl) event).generate(this.instant)) {
			synchronized (this.generatedEventList) {
				this.generatedEventList.add(event);
			}
			if (trace != null)
				trace.println(this + "." + this.instant + " " + reactiveEntity
						+ " generate " + event);
		}
	}

	/**
	 * Generates a MessageEvent to this community.
	 * @param reactiveEntity the sender ReactiveEntity.
	 * @param messageEvent the message Event to generate.
	 * @param message a Message.
	 */
	void generateMessage(ReactiveEntity reactiveEntity,
			Event messageEvent, Message message) {
		if (messageEvent.getCommunity() != this)
			throw new NotOwnEventException();
		MessageEventImpl messageEventI = (MessageEventImpl) messageEvent;
		if (messageEventI.generate(this.instant))
			if (messageEvent.isExclusive())
				synchronized (this.generatedExclusiveMessageEventList) {
					this.generatedExclusiveMessageEventList.add(messageEvent);
				}
			else
				synchronized (this.generatedMessageEventList) {
					this.generatedMessageEventList.add(messageEvent);
				}
		if (message != null)
			messageEventI.addGeneratedMessage(message);
		if (trace != null && (message == null || message.isTraceable()))
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " generate " + messageEvent + " " + message);
	}

	/**
	 * Generates an Action that concerns this community. The Action will be
	 * executed at the next end of instant.
	 * 
	 * @param reactiveEntity
	 *            the initiator entity.
	 * @param action
	 *            an Action.
	 */
	void generateAction(ReactiveEntity reactiveEntity, Action action) {
		if (trace != null && action.isTraceable())
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " generate " + action);
		synchronized (this.executableActionList) {
			this.executableActionList.add(action);
		}
	}
	
	List<Event> suspend(ReactiveEntityImpl reactiveEntity) {
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " suspend");
		EntityGroup entityGroup = reactiveEntity.getExclusiveEntityGroup();
		if (entityGroup == null)
			entityGroup = this.suspendedEntityGroup;
		entityGroup.oneEntityAchieveItsCycle(reactiveEntity);
		return this.unmodifiedBroadcastedEventList;
	}

	void stop(ReactiveEntityImpl reactiveEntity) {
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " stop");
		this.stoppedEntityGroup.oneEntityAchieveItsCycle(reactiveEntity);
	}

	void disconnect(ReactiveEntity reactiveEntity) {
		synchronized (this.connectedEntitySet) {
			this.connectedEntitySet.remove(reactiveEntity);
		}
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " disconnected");
		if (this.decrementActivityCount(reactiveEntity) == 0)
			this.startNewCycle();
	}

	EntityGroup getEntityGroup(ReactiveEntity exclusiveEntity) {
		return this.fromExclusiveEntityToEntityGroup.get(exclusiveEntity);
	}
	
	void putEntityGroup(ReactiveEntity exclusiveEntity, EntityGroup entityGroup) {
		this.fromExclusiveEntityToEntityGroup.put(exclusiveEntity, entityGroup);
	}
	
	void removeEntityGroup(ReactiveEntity exclusiveEntity) {
		this.fromExclusiveEntityToEntityGroup.remove(exclusiveEntity);
	}
	
	// /////////////////////////////////
	// ReactiveCommunity implementation
	// /////////////////////////////////

	public int getStatus() {
		return this.status;
	}

	public long getInstant() {
		return this.instant;
	}

	public Event createEvent() {
		return new EventImpl(this);
	}

	public Event createMessageEvent() {
		return new MessageEventImpl(this);
	}

	public Event getActionMessageEvent() {
		return this.actionMessageEvent;
	}

	public Board getBoard() {
		return this.board;
	}

	public void setBoard(Board board, long permissionKey) {
		if (this.futureBoard != null)
			throw new BoardAlreadyAttachedException();
		this.futureBoard = board;
		this.boardPermissionKey = permissionKey;
	}

	public String launchAgent(Agent agent) {
		return this.reactiveMachine.launchAgent(agent, this);
	}

	public int getConnectedEntityCount() {
		return this.connectedEntitySet.size();
	}
	
	// //////////////////////
	// Object implementation
	// //////////////////////

	public String toString() {
		return this.communityId;
	}
}
